use crate::{api::DynamicYtMusic, Command};
use std::borrow::Borrow;
use ytmapi_rs::{
    auth::{BrowserToken, OAuthToken},
    common::{
        AlbumID, ArtistChannelID, BrowseParams, EpisodeID, FeedbackTokenAddToLibrary,
        FeedbackTokenRemoveFromHistory, LikeStatus, MoodCategoryParams, PlaylistID,
        PodcastChannelID, PodcastChannelParams, PodcastID, SetVideoID, SongTrackingUrl, TasteToken,
        TasteTokenImpression, TasteTokenSelection, UploadAlbumID, UploadArtistID, UploadEntityID,
        VideoID, YoutubeID,
    },
    continuations::Continuable,
    parse::ParseFrom,
    process_json,
    query::{
        rate::{RatePlaylistQuery, RateSongQuery},
        song::GetSongTrackingUrlQuery,
        AddHistoryItemQuery, AddPlaylistItemsQuery, AlbumsFilter, ArtistsFilter,
        CommunityPlaylistsFilter, CreatePlaylistQuery, DeletePlaylistQuery,
        DeleteUploadEntityQuery, EditPlaylistQuery, EditSongLibraryStatusQuery, EpisodesFilter,
        FeaturedPlaylistsFilter, GetAlbumQuery, GetArtistAlbumsQuery, GetArtistQuery,
        GetChannelEpisodesQuery, GetChannelQuery, GetContinuationsQuery, GetEpisodeQuery,
        GetHistoryQuery, GetLibraryAlbumsQuery, GetLibraryArtistSubscriptionsQuery,
        GetLibraryArtistsQuery, GetLibraryPlaylistsQuery, GetLibrarySongsQuery,
        GetLibraryUploadAlbumQuery, GetLibraryUploadAlbumsQuery, GetLibraryUploadArtistQuery,
        GetLibraryUploadArtistsQuery, GetLibraryUploadSongsQuery, GetMoodCategoriesQuery,
        GetMoodPlaylistsQuery, GetNewEpisodesQuery, GetPlaylistQuery, GetPodcastQuery,
        GetSearchSuggestionsQuery, GetTasteProfileQuery, PlaylistsFilter, PodcastsFilter,
        PostQuery, ProfilesFilter, Query, RemoveHistoryItemsQuery, RemovePlaylistItemsQuery,
        SearchQuery, SetTasteProfileQuery, SongsFilter, VideosFilter,
    },
};

pub struct CliQuery {
    pub query_type: QueryType,
    pub show_source: bool,
}

pub enum QueryType {
    FromSourceFiles(Vec<String>),
    FromApi,
}

pub async fn command_to_query(
    command: Command,
    cli_query: CliQuery,
    yt: DynamicYtMusic,
) -> anyhow::Result<String> {
    match command {
        Command::GetSearchSuggestions { query } => {
            get_string_output_of_query(yt, GetSearchSuggestionsQuery::from(query), cli_query).await
        }
        Command::GetArtist { channel_id } => {
            get_string_output_of_query(
                yt,
                GetArtistQuery::new(ArtistChannelID::from_raw(channel_id)),
                cli_query,
            )
            .await
        }
        Command::GetPlaylist { playlist_id } => {
            get_string_output_of_query(
                yt,
                GetPlaylistQuery::new(PlaylistID::from_raw(playlist_id)),
                cli_query,
            )
            .await
        }
        Command::GetArtistAlbums {
            channel_id,
            browse_params,
        } => {
            get_string_output_of_query(
                yt,
                GetArtistAlbumsQuery::new(
                    ArtistChannelID::from_raw(channel_id),
                    BrowseParams::from_raw(browse_params),
                ),
                cli_query,
            )
            .await
        }
        Command::Search { query } => {
            get_string_output_of_query(yt, SearchQuery::new(query), cli_query).await
        }
        Command::SearchArtists { query } => {
            get_string_output_of_query(
                yt,
                SearchQuery::new(query).with_filter(ArtistsFilter),
                cli_query,
            )
            .await
        }
        Command::SearchAlbums { query } => {
            get_string_output_of_query(
                yt,
                SearchQuery::new(query).with_filter(AlbumsFilter),
                cli_query,
            )
            .await
        }
        Command::SearchSongs { query } => {
            get_string_output_of_query(
                yt,
                SearchQuery::new(query).with_filter(SongsFilter),
                cli_query,
            )
            .await
        }
        Command::SearchPlaylists { query } => {
            get_string_output_of_query(
                yt,
                SearchQuery::new(query).with_filter(PlaylistsFilter),
                cli_query,
            )
            .await
        }
        Command::SearchCommunityPlaylists { query } => {
            get_string_output_of_query(
                yt,
                SearchQuery::new(query).with_filter(CommunityPlaylistsFilter),
                cli_query,
            )
            .await
        }
        Command::SearchFeaturedPlaylists { query } => {
            get_string_output_of_query(
                yt,
                SearchQuery::new(query).with_filter(FeaturedPlaylistsFilter),
                cli_query,
            )
            .await
        }
        Command::SearchVideos { query } => {
            get_string_output_of_query(
                yt,
                SearchQuery::new(query).with_filter(VideosFilter),
                cli_query,
            )
            .await
        }
        Command::SearchEpisodes { query } => {
            get_string_output_of_query(
                yt,
                SearchQuery::new(query).with_filter(EpisodesFilter),
                cli_query,
            )
            .await
        }
        Command::SearchProfiles { query } => {
            get_string_output_of_query(
                yt,
                SearchQuery::new(query).with_filter(ProfilesFilter),
                cli_query,
            )
            .await
        }
        Command::SearchPodcasts { query } => {
            get_string_output_of_query(
                yt,
                SearchQuery::new(query).with_filter(PodcastsFilter),
                cli_query,
            )
            .await
        }
        Command::DeletePlaylist { playlist_id } => {
            get_string_output_of_query(
                yt,
                DeletePlaylistQuery::new(PlaylistID::from_raw(playlist_id)),
                cli_query,
            )
            .await
        }
        Command::GetAlbum { browse_id } => {
            get_string_output_of_query(
                yt,
                GetAlbumQuery::new(AlbumID::from_raw(browse_id)),
                cli_query,
            )
            .await
        }
        Command::CreatePlaylist { title, description } => {
            get_string_output_of_query(
                yt,
                CreatePlaylistQuery::new(
                    title.as_str(),
                    description.as_deref(),
                    Default::default(),
                ),
                cli_query,
            )
            .await
        }
        Command::RemovePlaylistItems {
            playlist_id,
            video_ids: set_video_ids,
        } => {
            get_string_output_of_query(
                yt,
                RemovePlaylistItemsQuery::new(
                    PlaylistID::from_raw(playlist_id),
                    set_video_ids.iter().map(SetVideoID::from_raw).collect(),
                ),
                cli_query,
            )
            .await
        }
        Command::AddVideosToPlaylist {
            playlist_id,
            video_ids,
        } => {
            get_string_output_of_query(
                yt,
                AddPlaylistItemsQuery::new_from_videos(
                    PlaylistID::from_raw(playlist_id),
                    video_ids.iter().map(VideoID::from_raw).collect(),
                    Default::default(),
                ),
                cli_query,
            )
            .await
        }
        Command::EditPlaylistTitle {
            playlist_id,
            new_title,
        } => {
            get_string_output_of_query(
                yt,
                EditPlaylistQuery::new_title(PlaylistID::from_raw(playlist_id), new_title),
                cli_query,
            )
            .await
        }
        Command::AddPlaylistToPlaylist {
            playlist_id,
            from_playlist_id,
        } => {
            get_string_output_of_query(
                yt,
                AddPlaylistItemsQuery::new_from_playlist(
                    PlaylistID::from_raw(playlist_id),
                    PlaylistID::from_raw(from_playlist_id),
                ),
                cli_query,
            )
            .await
        }
        Command::GetLibraryPlaylists { max_pages } => {
            get_string_output_of_streaming_query(yt, GetLibraryPlaylistsQuery, cli_query, max_pages)
                .await
        }
        Command::GetLibraryArtists { max_pages } => {
            get_string_output_of_streaming_query(
                yt,
                GetLibraryArtistsQuery::default(),
                cli_query,
                max_pages,
            )
            .await
        }
        Command::GetLibrarySongs { max_pages } => {
            get_string_output_of_streaming_query(
                yt,
                GetLibrarySongsQuery::default(),
                cli_query,
                max_pages,
            )
            .await
        }
        Command::GetLibraryAlbums { max_pages } => {
            get_string_output_of_streaming_query(
                yt,
                GetLibraryAlbumsQuery::default(),
                cli_query,
                max_pages,
            )
            .await
        }
        Command::GetLibraryArtistSubscriptions { max_pages } => {
            get_string_output_of_streaming_query(
                yt,
                GetLibraryArtistSubscriptionsQuery::default(),
                cli_query,
                max_pages,
            )
            .await
        }
        Command::GetHistory => get_string_output_of_query(yt, GetHistoryQuery, cli_query).await,
        Command::RemoveHistoryItems { feedback_tokens } => {
            get_string_output_of_query(
                yt,
                RemoveHistoryItemsQuery::new(
                    feedback_tokens
                        .iter()
                        .map(FeedbackTokenRemoveFromHistory::from_raw)
                        .collect(),
                ),
                cli_query,
            )
            .await
        }
        Command::RateSong {
            video_id,
            like_status,
        } => {
            get_string_output_of_query(
                yt,
                RateSongQuery::new(
                    VideoID::from_raw(video_id),
                    match like_status.as_str() {
                        "Like" => LikeStatus::Liked,
                        "Dislike" => LikeStatus::Disliked,
                        "Indifferent" => LikeStatus::Indifferent,
                        other => panic!("Unhandled like status <{other}>"),
                    },
                ),
                cli_query,
            )
            .await
        }
        Command::RatePlaylist {
            playlist_id,
            like_status,
        } => {
            get_string_output_of_query(
                yt,
                RatePlaylistQuery::new(
                    PlaylistID::from_raw(playlist_id),
                    match like_status.as_str() {
                        "Like" => LikeStatus::Liked,
                        "Dislike" => LikeStatus::Disliked,
                        "Indifferent" => LikeStatus::Indifferent,
                        other => panic!("Unhandled like status <{other}>"),
                    },
                ),
                cli_query,
            )
            .await
        }
        Command::EditSongLibraryStatus { feedback_tokens } => {
            get_string_output_of_query(
                yt,
                // Internal knowledge: Even though the string tokens we are provided could be
                // either Add or Remove tokens, it's OK to just provide
                // FeedBackTokenAddToLibrary's, as the tokens themselves determine if they will add
                // or remove.
                EditSongLibraryStatusQuery::new_from_add_to_library_feedback_tokens(
                    feedback_tokens
                        .iter()
                        .map(FeedbackTokenAddToLibrary::from_raw)
                        .collect(),
                ),
                cli_query,
            )
            .await
        }
        Command::GetLibraryUploadSongs => {
            get_string_output_of_query(yt, GetLibraryUploadSongsQuery::default(), cli_query).await
        }
        Command::GetLibraryUploadArtists => {
            get_string_output_of_query(yt, GetLibraryUploadArtistsQuery::default(), cli_query).await
        }
        Command::GetLibraryUploadAlbums => {
            get_string_output_of_query(yt, GetLibraryUploadAlbumsQuery::default(), cli_query).await
        }
        Command::GetLibraryUploadArtist { upload_artist_id } => {
            get_string_output_of_query(
                yt,
                GetLibraryUploadArtistQuery::new(UploadArtistID::from_raw(upload_artist_id)),
                cli_query,
            )
            .await
        }
        Command::GetLibraryUploadAlbum { upload_album_id } => {
            get_string_output_of_query(
                yt,
                GetLibraryUploadAlbumQuery::new(UploadAlbumID::from_raw(upload_album_id)),
                cli_query,
            )
            .await
        }
        Command::DeleteUploadEntity { upload_entity_id } => {
            get_string_output_of_query(
                yt,
                DeleteUploadEntityQuery::new(UploadEntityID::from_raw(upload_entity_id)),
                cli_query,
            )
            .await
        }
        Command::GetTasteProfile => {
            get_string_output_of_query(yt, GetTasteProfileQuery, cli_query).await
        }
        Command::SetTasteProfile {
            impression_token,
            selection_token,
        } => {
            get_string_output_of_query(
                yt,
                SetTasteProfileQuery::new([TasteToken {
                    impression_value: TasteTokenImpression::from_raw(impression_token),
                    selection_value: TasteTokenSelection::from_raw(selection_token),
                }]),
                cli_query,
            )
            .await
        }
        Command::GetMoodCategories => {
            get_string_output_of_query(yt, GetMoodCategoriesQuery, cli_query).await
        }
        Command::GetMoodPlaylists {
            mood_category_params,
        } => {
            get_string_output_of_query(
                yt,
                GetMoodPlaylistsQuery::new(MoodCategoryParams::from_raw(mood_category_params)),
                cli_query,
            )
            .await
        }
        Command::AddHistoryItem {
            song_tracking_url: song_url,
        } => {
            get_string_output_of_query(
                yt,
                AddHistoryItemQuery::new(SongTrackingUrl::from_raw(song_url)),
                cli_query,
            )
            .await
        }
        Command::GetSongTrackingUrl { video_id } => {
            get_string_output_of_query(
                yt,
                GetSongTrackingUrlQuery::new(VideoID::from_raw(video_id))?,
                cli_query,
            )
            .await
        }
        Command::GetChannel { channel_id } => {
            get_string_output_of_query(
                yt,
                GetChannelQuery::new(PodcastChannelID::from_raw(channel_id)),
                cli_query,
            )
            .await
        }
        Command::GetChannelEpisodes {
            channel_id,
            podcast_channel_params,
        } => {
            get_string_output_of_query(
                yt,
                GetChannelEpisodesQuery::new(
                    PodcastChannelID::from_raw(channel_id),
                    PodcastChannelParams::from_raw(podcast_channel_params),
                ),
                cli_query,
            )
            .await
        }
        Command::GetPodcast { podcast_id } => {
            get_string_output_of_query(
                yt,
                GetPodcastQuery::new(PodcastID::from_raw(podcast_id)),
                cli_query,
            )
            .await
        }
        Command::GetEpisode { video_id } => {
            get_string_output_of_query(
                yt,
                GetEpisodeQuery::new(EpisodeID::from_raw(video_id)),
                cli_query,
            )
            .await
        }
        Command::GetNewEpisodes => {
            get_string_output_of_query(yt, GetNewEpisodesQuery, cli_query).await
        }
    }
}

async fn get_string_output_of_query<Q, O>(
    yt: DynamicYtMusic,
    q: impl Borrow<Q>,
    cli_query: CliQuery,
) -> anyhow::Result<String>
where
    Q: Query<BrowserToken, Output = O>,
    Q: Query<OAuthToken, Output = O>,
    O: ParseFrom<Q>,
{
    match cli_query {
        CliQuery {
            query_type: QueryType::FromApi,
            show_source: true,
        } => yt.query_source(q.borrow()).await,
        CliQuery {
            query_type: QueryType::FromApi,
            show_source: false,
        } => yt.query(q).await.map(|r| format!("{:#?}", r)),
        CliQuery {
            query_type: QueryType::FromSourceFiles(sources),
            show_source: true,
        } => Ok(sources.into_iter().next().unwrap_or_default()),
        CliQuery {
            query_type: QueryType::FromSourceFiles(sources),
            show_source: false,
        } => {
            // Note - if multiple sources are provided, only the first is processed - the
            // rest are ignored.
            if let Some(first_source) = sources.into_iter().next() {
                process_json_based_on_dyn_api(&yt, first_source, q)
            } else {
                Ok(String::new())
            }
        }
    }
}

async fn get_string_output_of_streaming_query<Q, O>(
    yt: DynamicYtMusic,
    q: impl Borrow<Q>,
    cli_query: CliQuery,
    max_pages: usize,
) -> anyhow::Result<String>
where
    Q: Query<BrowserToken, Output = O>,
    Q: Query<OAuthToken, Output = O>,
    Q: PostQuery,
    O: ParseFrom<Q> + Continuable<Q>,
{
    match cli_query {
        CliQuery {
            query_type: QueryType::FromApi,
            show_source: true,
        } => yt.stream_source(q.borrow(), max_pages).await,
        CliQuery {
            query_type: QueryType::FromApi,
            show_source: false,
        } => yt.stream(q, max_pages).await.map(|r| format!("{:#?}", r)),
        CliQuery {
            query_type: QueryType::FromSourceFiles(sources),
            show_source: true,
        } => {
            Ok(
                // Replace with standard library method once stabilised.
                itertools::intersperse(sources.into_iter().take(max_pages), "\n".to_string())
                    .collect(),
            )
        }
        CliQuery {
            query_type: QueryType::FromSourceFiles(sources),
            show_source: false,
        } => {
            let mut output_arr = vec![];
            let mut sources_iter = sources.into_iter().take(max_pages);
            if let Some(first_source) = sources_iter.next() {
                output_arr.push(process_json_based_on_dyn_api::<Q, O>(
                    &yt,
                    first_source,
                    q.borrow(),
                )?)
            }
            for source in sources_iter {
                let continuation_query = GetContinuationsQuery::new_mock_unchecked(q.borrow());
                output_arr.push(process_json_based_on_dyn_api(
                    &yt,
                    source,
                    continuation_query,
                )?)
            }
            Ok(output_arr.join("\n"))
        }
    }
}

fn process_json_based_on_dyn_api<Q, O>(
    yt: &DynamicYtMusic,
    source: String,
    query: impl Borrow<Q>,
) -> anyhow::Result<String>
where
    Q: Query<BrowserToken, Output = O>,
    Q: Query<OAuthToken, Output = O>,
    O: ParseFrom<Q>,
{
    // The matching on yt is a neat hack to ensure process_json utilises the same
    // AuthType as was set in config. This works as the config step sets
    // the variant of DynamicYtMusic.
    match yt {
        DynamicYtMusic::Browser(_) => process_json::<Q, BrowserToken>(source, query)
            .map(|r| format!("{:#?}", r))
            .map_err(|e| e.into()),
        DynamicYtMusic::OAuth(_) => process_json::<Q, OAuthToken>(source, query)
            .map(|r| format!("{:#?}", r))
            .map_err(|e| e.into()),
    }
}
